module casper.parser {
   create =
      (FilenameRef = data.ref.backtrackable:new('')) >>
   module [FilenameRef] {
      filename = FilenameRef.

      term('callable'(F, 'escaped', 'none', Pb, Pe)) =
         ! >>
         'position'('string'(F), 'file'(Pb, Pe)).
      term('callable'(F, _, Args, Pb, Pe)) =
         'position'(
            data.term:make(
               F,
               (Args = 'none' -> []; data.list:iter(Args) { do(X) = term(X). })
            ),
            'file'(Pb, Pe)
         ).
      term('()'(Sub, _, _)) = term(Sub).
      term('.()'(_, _, _, _), _) :-
         throw('error'('.()', 'term')).
      term('module'(_, _, _, _, _, _, _), _) :-
         throw('error'('module', 'term')).
      term('number'(N, _, _)) = N.
      term('empty'(_, _), _) :-
         throw('error'('empty', 'term')).
      term('doc'(_, _, _), _) :-
         throw('error'('doc', 'term')).

      parse_stream(Stream, Ast) :-
         Filename = filename:get,
         Lexems = emop.lexer:utf8_stream(Stream, Filename),
         casper.new_parser:parse(Lexems, 'file', _, Ast0),
         Ast = term(emop.parse:reparse(Ast0) {
            syntax: infix left associative ':' '+' '-' '*' '/' 'mod' 'quo'.
            syntax: infix right associative ',' '->' '\\/' '/\\' ':=' '!=' '<=' '<-' 'in' '..' '**' '|' ''.
            syntax: prefix '+' '-' 'let' 'forall'.

            syntax: '|' <  'let' = 'forall' < ',' < '->' < ':' < '\\/' < '/\\' <  'in' < ':=' = '!=' = '<-' = '<=' < '..' < infix '+' = infix '-' < '*' = '/' = 'mod' = 'quo'.
            syntax: '*' < '**' < prefix '+' = prefix '-' < ''.
         }).
   }.


   parse_file(Filename, Ast) :-
      Stream = io.file:open_read(Filename),
      Parser = create,
      Parser:filename:set(Filename),
      Ast = Parser:parse_stream(Stream),
      Stream:close.
}